/*
 * pwd.c
 * Implementation of SUSv3 XCU pwd utility
 * Copyright © 2007 Rich Felker
 * Licensed under the terms of the GNU General Public License, v2 or later
 */

/* NOTE: This implementation assumes PATH_MAX is defined and correct. */

#include <limits.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

static int my_write(int fd, const char *s, size_t l)
{
	if (!l) l = strlen(s);
	while (l) {
		ssize_t n = write(fd, s, l);
		if (n<0) return -1;
		s += n; l -= n;
	}
	return 0;
}

static void my_perror(char *prog, char *msg)
{
	char *err = strerror(errno);
	write(2, prog, strlen(prog));
	write(2, ": ", 2);
	write(2, msg, strlen(msg));
	write(2, ": ", 2);
	write(2, err, strlen(err));
	write(2, "\n", 1);
}

static int is_rel(const char *s)
{
	unsigned dots=0, slash=1;
	if (*s++ != '/') return 1;
	for (; *s; s++) {
		if (*s == '/') {
			if (dots-1<2) return 1;
			dots=0;
		}
		else slash=1;
		if (slash && *s == '.') dots++;
		else dots=slash=0;
	}
	return dots-1<2;
}

int main(int argc, char *argv[])
{
	int i, j;
	int p=0;
	char buf[PATH_MAX+2];
	struct stat st1, st2;
	char *pwd;
	
	for (i=1; i<argc && argv[i][0]=='-'; i++)
		for (j=1; argv[i][j]; j++) {
			switch (argv[i][j]) {
			case 'L': p=0; continue;
			case 'P': p=1;
			case 0: continue;
			}
invalid:
			errno = EINVAL;
			my_perror(argv[0], argv[i]);
			return 1;
		}
	if (i<argc) goto invalid;
	if (p || !(pwd=getenv("PWD")) || is_rel(pwd)
	 || stat(pwd, &st1) || stat(".", &st2)
	 || st1.st_dev != st2.st_dev
	 || st1.st_ino != st2.st_ino) {
		if (!getcwd(buf, sizeof buf)) {
			my_perror(argv[0], "getcwd");
			return 1;
		}
	} else {
		/* safe because stat succeeded */
		strcpy(buf, pwd);
	}
	strcat(buf, "\n");
	return my_write(1, buf, 0) != 0;
}
